/*
 * Copyright (C) 2002 2003 2004, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


/* This piece of code is imported several times into datasource.c. */

/* Macros that must be defined and are undefined at the end of this file:
 * STYPE - Floating point sample type
 * CPF - convert pcm to fp function name
 * CPFA - convert pcm array to fp function name
 * CFP - convert fp to pcm function name
 * CFPA - convert fp array to pcm function name
 */

STYPE CPF(void *samplebuf, Dataformat *format)
{
     STYPE s;
     gint32 i32;
     guint32 u32;
     if (format->sign)
	  switch (format->samplesize) {
	  case 1: 
	       s=(STYPE)(*((gchar *)samplebuf));
	       s /= 128.0;
	       return s;
	  case 2: 
	       s=(STYPE)(*((gint16 *)samplebuf));
	       s /= 32768.0;
	       return s;
	  case 3: 
	       i32 = (*((gint32 *)samplebuf));
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
	       i32 &= 0xFFFFFF;
#else
	       i32 = (i32 & 0xFFFFFF00) >> 8;
#endif
	       if (i32 & 0x800000) i32 -= 16777216;
	       s = (STYPE) i32;
	       s/=8388608.0;
	       return s;
	  case 4: 
	       s=(STYPE)(*((gint32 *)samplebuf));
	       s/=2147483648.0;
	       return s;
	  }
     else 
	  switch (format->samplesize) {
	  case 1: 
               s=(STYPE)(*((guchar *)samplebuf));
               s-=127.5;
               s/=127.5;
               return s;
	  case 2: 
	       s=(STYPE)(*((guint16 *)samplebuf));
	       s-=32767.5;
	       s/=32767.5;
	       return s;
	  case 3: 
	       u32 = (*((guint32 *)samplebuf));
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
	       u32 &= 0xFFFFFF;
#else
	       u32 = (u32 & 0xFFFFFF00) > 8;
#endif
	       s = (STYPE) u32;
	       s-=8388607.5;
	       s/=8388607.5;
	       return s;
	  case 4: 
	       s=(STYPE)(*((guint32 *)samplebuf));
	       s-=2147483647.5;
	       s/=2147483647.5;
	       return s;
	  }
     g_assert_not_reached();
     return 0.0;
}

void CPFA(void *samplebuf, Dataformat *format, 
	  STYPE *buffer, guint count)
{
     STYPE s;
     gchar *c = samplebuf;
     gint32 i32;
     guint32 u32;
     if (format->sign)
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
		    s=(STYPE)(*((gchar *)c));
		    s/=128.0;
		    *buffer = s;
		    c += 1;
		    buffer ++;
	       }
	       break;
	  case 2: 
	       for (; count>0; count--) {
		    s=(STYPE)(*((gint16 *)c));
		    s/=32768.0;
		    *buffer = s;
		    c += 2;
		    buffer ++;
	       }
	       break;
	  case 3: 
	       for (; count>0; count--) {
		    i32 = (*((gint32 *)c));
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
		    i32 &= 0xFFFFFF;
#else
		    i32 = (i32 & 0xFFFFFF00) >> 8;
#endif
		    if (i32 & 0x800000) i32 -= 16777216;
		    s = (STYPE) i32;
		    s/=8388608.0;
		    *buffer = s;
		    c += 3;
		    buffer ++;
	       }
	       break;
	  case 4: 
	       for (; count>0; count--) {
		    s=(STYPE)(*((gint32 *)c));
		    s/=2147483648.0;
		    *buffer = s;
		    c += 4;
		    buffer ++;
	       }
	       break;
	  }
     else 
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
                    s=(STYPE)(*((guchar *)c));
                    s-=127.5;
                    s/=127.5;
                    *buffer = s;
                    c += 1;
                    buffer ++;
	       }
	       break;
	  case 2: 
	       for (; count>0; count--) {
		    s=(STYPE)(*((guint16 *)c));
		    s-=32767.5;
		    s/=32767.5;
		    *buffer = s;
		    c += 2;
		    buffer ++;
	       }
	       break;
	  case 3: 
	       for (; count>0; count--) {
		    u32 = (*((guint32 *)c));
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
		    u32 &= 0xFFFFFF;
#else
		    u32 = (u32 & 0xFFFFFF00) >> 8;
#endif
		    s = (STYPE) u32;
		    s-=8388607.5;
		    s/=8388607.5;
		    *buffer = s;
		    c += 3;
		    buffer ++;
	       }
	       break;
	  case 4: 
	       for (; count>0; count--) {
		    s=(STYPE)(*((guint32 *)c));
		    s-=2147483647.5;
		    s/=2147483647.5;
		    *buffer = s;
		    c += 4;
		    buffer ++;
	       }
	       break;
	  }
}

void CFP(STYPE d, void *samplebuf, Dataformat *format)
{
     gint32 i32;
     guint32 u32;
     guint8 u80,u81,u82;
     if (d < -1.0) d = -1.0;
     if (format->sign)
	  switch (format->samplesize) {
	  case 1: 
	       d*=128.0;
	       if (d > 127.0) d = 127.0;
	       (*((gchar *)samplebuf)) = (gchar)d;
	       return;
	  case 2: 
	       d*=32768.0;
	       if (d > 32767.0) d = 32767.0;
	       (*((gint16 *)samplebuf)) = (gint16)d;
	       return;
	  case 3: 
	       d*=8388608.0; 
	       if (d > 8388607.0) d = 8388607.0;
	       i32 = (gint32) d;
	       u32 = (guint32) i32;
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
	       u80 = u32;
	       u81 = u32 >> 8;
	       u82 = u32 >> 16;
#else
	       u80 = u32 >> 16;
	       u81 = u32 >> 8;
	       u82 = u32;
#endif
	       ((gint8 *)samplebuf)[0] = u80;
	       ((gint8 *)samplebuf)[1] = u81;
	       ((gint8 *)samplebuf)[2] = u82;
	       return;
	  case 4: 
	       d*=2147483648.0; 
	       if ( d > 2147483647.0) d = 2147483647.0;
	       (*((gint32 *)samplebuf)) = (gint32)d;
	       return;
	  }
     else {
	  if (d > ((STYPE)1.0)) d = ((STYPE)1.0);
	  switch (format->samplesize) {
	  case 1: 
               d*=127.5;
               d+=127.5;
               (*((guchar *)samplebuf)) = (guchar)d;
               return;
	  case 2: 
	       d*=32767.5;
	       d+=32767.5;
	       (*((guint16 *)samplebuf)) = (guint16)d;
	       return;
	  case 3: 
	       d*=8388607.5;	       
	       d+=8388607.5;	       
	       u32 = (guint32) d;
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
	       u80 = u32;
	       u81 = u32 >> 8;
	       u82 = u32 >> 16;
#else
	       u80 = u32 >> 16;
	       u81 = u32 >> 8;
	       u82 = u32;
#endif
	       ((gint8 *)samplebuf)[0] = u80;
	       ((gint8 *)samplebuf)[1] = u81;
	       ((gint8 *)samplebuf)[2] = u82;
	       return;
	  case 4: 
	       d*=2147483647.5;
	       d+=2147483647.5;
	       (*((guint32 *)samplebuf)) = (guint32)d;
	       return ;
	  }
     }
}

void CFPA(STYPE *inbuf, void *outbuf, Dataformat *format, 
	  guint count)
{
     gint32 i32;
     guint32 u32;
     guint8 u80,u81,u82;
     STYPE d,max;
     STYPE *in = inbuf;
     gchar *out = outbuf;
     if (format->sign) {
	  max = maximum_float_value(format);
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > max) d = max;
		    else if (d < -1.0) d = -1.0;
		    d*= ((STYPE)128.0);
		    (*((gchar *)out)) = (gchar)d;
		    in++;
		    out+=1;
	       }
	       return;
	  case 2: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > max) d = max;
		    else if (d < -1.0) d = -1.0;

		    d*= ((STYPE)32768.0);
		    (*((gint16 *)out)) = (gint16)d;
		    in ++;
		    out += 2;
	       }
	       return;
	  case 3: 
	       if (count == 0) return;
	       count--;
	       for (; count>0; count--) {
		    d = *in;
		    if (d > max) d = max;
		    else if (d < -1.0) d = -1.0;
		    d*=((STYPE)8388608.0);
		    i32 = (gint32) d;
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
		    (*((gint32 *)out)) = i32;
#else
		    (*((gint32 *)out)) = i32 << 8;
#endif
		    in ++;
		    out += 3;
	       }
	       /* The last sample has to be handled separately since
		* we don't want to touch the first byte after the output
		* buffer ends. */
	       d = *in;
	       if (d > max) d = 1.0;
	       else if (d < -1.0) d = -1.0;
	       d*=((STYPE)8388608.0);
	       i32 = (gint32) d;
	       u32 = (guint32) i32;
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
	       u80 = u32;
	       u81 = u32 >> 8;
	       u82 = u32 >> 16;
#else
	       u80 = u32 >> 16;
	       u81 = u32 >> 8;
	       u82 = u32;
#endif
	       out[0] = u80;
	       out[1] = u81;
	       out[2] = u82;
	       return;
	  case 4: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > max) d = max;
		    else if (d < -1.0) d = -1.0;
		    d*=((STYPE)2147483648.0); 
		    (*((gint32 *)out)) = (gint32)d;
		    in ++;
		    out += 4;
	       }
	       return;
	  }
     } else 
	  switch (format->samplesize) {
	  case 1: 
	       for (; count>0; count--) {
                    d = *in;
                    if (d > 1.0) d = 1.0;
                    else if (d < -1.0) d = -1.0;
                    d*=127.5;
                    d+=127.5;
                    (*((guchar *)out)) = (guchar)d;
                    in ++;
                    out += 1;
	       }
	       return;
	  case 2: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    d*=32767.5;
		    d+=32767.5;
		    (*((guint16 *)out)) = (guint16)d;
		    in ++;
		    out += 2;
	       }
	       return;
	  case 3: 
	       if (count == 0) return;
	       count --;
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    d*=8388607.5;	       
		    d+=8388607.5;	       
		    u32 = (guint32) d;
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
		    (*((guint32 *)out)) = u32;
#else
		    (*((guint32 *)out)) = u32 >> 8;
#endif
		    in ++;
		    out += 3;
	       }
	       d = *in;
	       if (d > 1.0) d = 1.0;
	       else if (d < -1.0) d = -1.0;
	       d*=8388607.5;	       
	       d+=8388607.5;	       
	       u32 = (guint32) d;
#if G_BYTE_ORDER == G_LITTLE_ENDIAN
	       u80 = u32;
	       u81 = u32 >> 8;
	       u82 = u32 >> 16;
#else
	       u80 = u32 >> 16;
	       u81 = u32 >> 8;
	       u82 = u32;
#endif
	       out[0] = u80;
	       out[1] = u81;
	       out[2] = u82;
	       return;
	  case 4: 
	       for (; count>0; count--) {
		    d = *in;
		    if (d > 1.0) d = 1.0;
		    else if (d < -1.0) d = -1.0;
		    d*=2147483647.5;
		    d+=2147483647.5;
		    (*((guint32 *)out)) = (guint32)d;
		    in ++;
		    out += 4;
	       }
	       return ;
	  }
}

#undef STYPE
#undef CPF
#undef CPFA
#undef CFP
#undef CFPA
