/*
 * Copyright (C) 2002 2003 2004, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include <config.h>

#include "main.h"
#include "formatselector.h"
#include "inifile.h"

static void format_selector_class_init(GtkObjectClass *klass)
{     
}

static void format_selector_toggle(GtkToggleButton *button, gpointer user_data)
{
     FormatSelector *fs = FORMAT_SELECTOR(user_data);
     gint i;
     if (!gtk_toggle_button_get_active(button)) return;
     for (i=1; i<5; i++)
	  if (GTK_WIDGET(button) == fs->samplesize_widgets[i-1])
	       fs->samplesize = i;
     for (i=0; i<2; i++)
	  if (GTK_WIDGET(button) == fs->sign_widgets[i])
	       fs->sign = (gboolean)i;
     for (i=0; i<2; i++)
	  if (GTK_WIDGET(button) == fs->endian_widgets[i])
	       fs->bigendian = (gboolean)i;     
}

static void format_selector_init(GtkWidget *widget)
{
     GtkBox *hbox = GTK_BOX(widget);
     FormatSelector *fs = FORMAT_SELECTOR(widget);
     GtkBox *vbox;
     GtkWidget *w;
     GtkRadioButton *group=NULL;
     gchar c[8];
     gint i;

     vbox = GTK_BOX(gtk_vbox_new(FALSE,3));

     for (i=1; i<5; i++) {
	  g_snprintf(c,sizeof(c),"%d-bit",i*8);
	  w = gtk_radio_button_new_with_label_from_widget (group, c);
	  gtk_signal_connect(GTK_OBJECT(w),"toggled",
			     GTK_SIGNAL_FUNC(format_selector_toggle),fs);
	  if (!group) group = GTK_RADIO_BUTTON(w);
	  gtk_box_pack_start(GTK_BOX(vbox),w,FALSE,FALSE,0);
	  gtk_widget_show(w);
	  fs->samplesize_widgets[i-1] = w;
     }

     gtk_box_pack_start(hbox,GTK_WIDGET(vbox),TRUE,TRUE,2);
     gtk_widget_show(GTK_WIDGET(vbox));
     
     w = gtk_vseparator_new ();
     gtk_box_pack_start(hbox,w,FALSE,FALSE,2);
     gtk_widget_show(w);

     vbox = GTK_BOX(gtk_vbox_new(FALSE,3));

     group = NULL;
     for (i=0;i<2;i++) {
	  w = gtk_radio_button_new_with_label_from_widget ( 
	       group, i?"Signed":"Unsigned" );
	  gtk_signal_connect(GTK_OBJECT(w),"toggled",
			     GTK_SIGNAL_FUNC(format_selector_toggle),fs);
	  if (!group) group = GTK_RADIO_BUTTON(w);
	  gtk_box_pack_start(vbox,w,FALSE,FALSE,0);
	  gtk_widget_show(w);
	  fs->sign_widgets[i] = w;
     }

     w = gtk_hseparator_new();
     gtk_box_pack_start(vbox,w,FALSE,FALSE,0);
     gtk_widget_show(w);
     
     group = NULL;
     for (i=0;i<2;i++) {
	  w = gtk_radio_button_new_with_label_from_widget(
	       group,(i==0)?"Little-endian":"Big-endian");
	  gtk_signal_connect(GTK_OBJECT(w),"toggled",
			     GTK_SIGNAL_FUNC(format_selector_toggle),fs);
	  if (!group) group = GTK_RADIO_BUTTON(w);
	  gtk_box_pack_start(vbox,GTK_WIDGET(w),FALSE,FALSE,0);
	  gtk_widget_show(GTK_WIDGET(w));
	  fs->endian_widgets[i] = w;
     }

     gtk_box_pack_start(hbox,GTK_WIDGET(vbox),TRUE,TRUE,2);
     gtk_widget_show(GTK_WIDGET(vbox));

     format_selector_set(fs,1,FALSE,FALSE);
}

GtkType format_selector_get_type(void)
{
     static GtkType id = 0;
     if (!id) {
	  GtkTypeInfo info = {
	       "FormatSelector",
	       sizeof(FormatSelector),
	       sizeof(FormatSelectorClass),
	       (GtkClassInitFunc) format_selector_class_init,
	       (GtkObjectInitFunc) format_selector_init
	  };
	  id = gtk_type_unique(gtk_hbox_get_type(),&info);
     }
     return id;
}

GtkWidget *format_selector_new(void)
{
     return gtk_type_new(format_selector_get_type());
}

void format_selector_set(FormatSelector *fs, guint samplesize, gboolean sign,
			 gboolean bigendian)
{
     if (samplesize <= 4 && samplesize >= 1) {
	  gtk_toggle_button_set_active(
	       GTK_TOGGLE_BUTTON(fs->samplesize_widgets[samplesize-1]),TRUE);
	  fs->samplesize = samplesize;
     }
     gtk_toggle_button_set_active(
	  GTK_TOGGLE_BUTTON(fs->sign_widgets[sign?1:0]),TRUE);
     fs->sign = sign;
     gtk_toggle_button_set_active(
	  GTK_TOGGLE_BUTTON(fs->endian_widgets[bigendian?1:0]),TRUE);
     fs->bigendian = bigendian;
}

void format_selector_set_from_inifile(FormatSelector *fs, gchar *ini_prefix)
{
     guint32 ss;
     gboolean sn,be;
     gchar *c;
     int l;
     l = strlen(ini_prefix);
     c = g_malloc(l+12);
     strcpy(c,ini_prefix);
     strcpy(c+l,"_SampleSize");
     ss = inifile_get_guint32(c,fs->samplesize);
     strcpy(c+l,"_Signed");
     sn = inifile_get_gboolean(c,fs->sign);
     strcpy(c+l,"_BigEndian");
     be = inifile_get_gboolean(c,fs->bigendian);
     format_selector_set(fs,ss,sn,be);
     g_free(c);
}

void format_selector_save_to_inifile(FormatSelector *fs, gchar *ini_prefix)
{
     gchar *c;
     int l;
     l = strlen(ini_prefix);
     c = g_malloc(l+12);
     strcpy(c,ini_prefix);
     strcpy(c+l,"_SampleSize");
     inifile_set_guint32(c,fs->samplesize);
     strcpy(c+l,"_Signed");
     inifile_set_gboolean(c,fs->sign);
     strcpy(c+l,"_BigEndian");
     inifile_set_gboolean(c,fs->bigendian);
     g_free(c);
}

void format_selector_show_endian(FormatSelector *fs, gboolean visible)
{
     if (visible) {
	  gtk_widget_show(fs->endian_widgets[0]);
	  gtk_widget_show(fs->endian_widgets[1]);
     } else {
	  gtk_widget_hide(fs->endian_widgets[0]);
	  gtk_widget_hide(fs->endian_widgets[1]);
     }
}
