/*
 * Copyright (C) 2002 2003 2004, Magnus Hjorth
 *
 * This file is part of mhWaveEdit.
 *
 * mhWaveEdit is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by        
 * the Free Software Foundation; either version 2 of the License, or  
 * (at your option) any later version.
 *
 * mhWaveEdit is distributed in the hope that it will be useful,   
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with mhWaveEdit; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 */


#include "mainwindowlist.h"

enum { CHANGED_SIGNAL, LAST_SIGNAL };
static gint mainwindow_list_signals[LAST_SIGNAL] = { 0 };
static gboolean updating = FALSE;

static void mainwindow_list_entry_changed(GtkEditable *editable, 
					  gpointer user_data)
{
     gint i,s;
     Mainwindow *w=NULL;
     gchar *c;
     MainwindowList *mwl = MAINWINDOW_LIST(user_data);

     if (updating) return;
     c = (gchar *)gtk_entry_get_text(GTK_ENTRY(editable));
     if (c[0] == 0) return;
     s = list_object_get_size(mainwindow_objects);
     for (i=0; i<s; i++) {
	  w = MAINWINDOW(list_object_get(mainwindow_objects,i));
	  if (w->titlename!=NULL && !strcmp(w->titlename,c)) break;
     }
     g_assert(i<s);
     if (w != mwl->selected) {
	  mwl->selected = w;
	  gtk_signal_emit(GTK_OBJECT(mwl),
			  mainwindow_list_signals[CHANGED_SIGNAL]);
     }
}

static void mainwindow_list_addnotify(ListObject *lo, gpointer item, 
				      gpointer user_data)
{
     Mainwindow *w = MAINWINDOW(item);
     MainwindowList *mwl = MAINWINDOW_LIST(user_data);
     if (mwl->selected == NULL) {
	  mwl->selected=w;
	  gtk_widget_set_sensitive(GTK_WIDGET(mwl),TRUE);
     }
     mainwindow_list_setup(mwl,mwl->selected);     
     if (w == mwl->selected) {	  
	  gtk_signal_emit(GTK_OBJECT(mwl),
			  mainwindow_list_signals[CHANGED_SIGNAL]);
     }
}

static void mainwindow_list_remove(ListObject *lo, gpointer item,
				   gpointer user_data)
{
     Mainwindow *w = MAINWINDOW(item);
     MainwindowList *mwl = MAINWINDOW_LIST(user_data);
     if (w == mwl->selected) {
	  if (list_object_get_size(mainwindow_objects) == 0) {
	       /* We set the selected item to NULL and dim the widget, but 
		* no signal is sent out. */
	       mwl->selected = NULL;
	       gtk_widget_set_sensitive(GTK_WIDGET(mwl),FALSE);
	       return;
	  }
	  mwl->selected = MAINWINDOW(list_object_get(mainwindow_objects,0));
	  mainwindow_list_setup(mwl,mwl->selected);
	  gtk_signal_emit(GTK_OBJECT(mwl),
			  mainwindow_list_signals[CHANGED_SIGNAL]);
     } else
	  mainwindow_list_setup(mwl,mwl->selected);

}

static void mainwindow_list_init(GtkObject *obj)
{
     MainwindowList *mwl = MAINWINDOW_LIST(obj);
     /* Most initialization is done in mainwindow_list_setup */
     mwl->selected = NULL;
     gtk_editable_set_editable(GTK_EDITABLE(GTK_COMBO(obj)->entry),FALSE);
     gtk_signal_connect(GTK_OBJECT(GTK_COMBO(obj)->entry),"changed",
			GTK_SIGNAL_FUNC(mainwindow_list_entry_changed),obj);
     gtk_signal_connect_while_alive(GTK_OBJECT(mainwindow_objects), 
				    "item_added",
				    GTK_SIGNAL_FUNC(mainwindow_list_addnotify),
				    obj,obj);
     gtk_signal_connect_while_alive(GTK_OBJECT(mainwindow_objects), 
				    "item_removed",
				    GTK_SIGNAL_FUNC(mainwindow_list_remove),
				    obj,obj);
     gtk_signal_connect_while_alive(GTK_OBJECT(mainwindow_objects), 
				    "item_notify",
				    GTK_SIGNAL_FUNC(mainwindow_list_addnotify),
				    obj,obj);
}

static void mainwindow_list_class_init(GtkObjectClass *klass)
{
     MainwindowListClass *mwlc = MAINWINDOW_LIST_CLASS(klass);
     mwlc->changed = NULL;
     mainwindow_list_signals[CHANGED_SIGNAL] = 
	  gtk_signal_new( "changed", GTK_RUN_FIRST, GTK_CLASS_TYPE(klass),
			  GTK_SIGNAL_OFFSET(MainwindowListClass,changed),
			  gtk_marshal_NONE__NONE, GTK_TYPE_NONE, 0);
     gtk_object_class_add_signals(klass,mainwindow_list_signals,LAST_SIGNAL);
}

GtkType mainwindow_list_get_type(void)
{
     static GtkType id=0;
     if (!id) {
	  GtkTypeInfo info = {
	       "MainwindowList",
	       sizeof(MainwindowList),
	       sizeof(MainwindowListClass),
	       (GtkClassInitFunc)mainwindow_list_class_init,
	       (GtkObjectInitFunc)mainwindow_list_init
	  };
	  id = gtk_type_unique( gtk_combo_get_type(), &info );	       
     }	
     return id;
}

GtkWidget *mainwindow_list_new(Mainwindow *chosen)
{
     GtkWidget *widget;
     widget = GTK_WIDGET(gtk_type_new(mainwindow_list_get_type()));
     mainwindow_list_setup(MAINWINDOW_LIST(widget),chosen);
     return widget;     
}

static void mainwindow_list_setup_func(gpointer item, gpointer user_data)
{
     GList **lp = (GList **)user_data;
     Mainwindow *w = MAINWINDOW(item);
     if (w->titlename != NULL)
	  *lp = g_list_append(*lp, w->titlename);
}

void mainwindow_list_setup(MainwindowList *mwl, Mainwindow *chosen)
{   
     GList *l = NULL;
     updating = TRUE;     
     gtk_list_clear_items(GTK_LIST(GTK_COMBO(mwl)->list),0,-1);
     list_object_foreach(mainwindow_objects,mainwindow_list_setup_func,&l);
     gtk_combo_set_popdown_strings(GTK_COMBO(mwl),l);
     g_list_free(l);     
     if (chosen) {
	  gtk_entry_set_text(GTK_ENTRY(GTK_COMBO(mwl)->entry),
			     chosen->titlename);
	  memcpy(&(mwl->format),&(chosen->view->chunk->format),
		 sizeof(Dataformat));
	  mwl->selected = chosen;
     }
     updating = FALSE;
}
